﻿#include <stdio.h>
#include <stdlib.h>
#include "fisheyesw.h"
#include "SDL/SDL.h"


#ifndef _LINUX
#pragma comment(lib, "SDLmain.lib")
#pragma comment(lib, "SDL.lib")
#pragma comment(lib, "d_fisheyesw.lib")
#endif

#ifdef _LINUX
#include <string.h>
#endif


#define _MULTI_VIEW

#ifdef _MULTI_VIEW
#define MULTI_VIEW_COUNT 4
typedef struct _tCustomControlOption
{
	long	Pitch;
	long	Roll;
	FERECT	OutRoi; 
}TCUSTOMCTRLOPTION;
static TCUSTOMCTRLOPTION gs_apCtrlOption[MULTI_VIEW_COUNT];
#endif  // _MULTI_VIEW

void EventLoop(HANDLE hObject, FEOPTION *pOption, SDL_Overlay *pOverlay);
SCODE UpdateOutVPicture(HANDLE hObject, FEOPTION *pOption, SDL_Overlay *pOverlay); 

int main(int argc, char *argv[])
{
	const int width = 960, height = 540, bpp = 32;
	const DWORD dwFovSize = 1536;
	FILE *pfSample = NULL;
	char *pszimgPath;
	int mountType = 0;
	int sdlRet;
	SDL_Surface *pScreen;
	SDL_Overlay *pOverlay;
	SCODE scRet;
	HANDLE hFECtx;
	FEOPTION option;
	
	if (argc < 3)
	{
		printf("invalid parameters!\n");
		return -1;
	}

	// Get user-specified image path and mount type
	pszimgPath = argv[1];
	mountType = atoi(argv[2]);

	if (NULL == (pfSample = fopen(pszimgPath, "rb")))
	{
		return -1;
	}
	
	sdlRet = SDL_Init(SDL_INIT_VIDEO);
	if (sdlRet)
	{
		return sdlRet;
	}

	if (NULL == (pScreen = SDL_SetVideoMode(width, height, bpp, SDL_DOUBLEBUF)))
	{
		return -1;
	}
	
	if (NULL == (pOverlay = SDL_CreateYUVOverlay(width, height, SDL_YV12_OVERLAY, pScreen)))
	{
		return -1;
	}

	// Initial Fisheye library
	scRet = Fisheye_Initial(&hFECtx, LIBFISHEYE_VERSION);
	if (S_OK != scRet)
	{
		return -1;
	}

	memset(&option, 0, sizeof(FEOPTION));
	option.Scale = 100;

#ifdef _MULTI_VIEW
	memset(gs_apCtrlOption, 0, sizeof(TCUSTOMCTRLOPTION) * MULTI_VIEW_COUNT);
	gs_apCtrlOption[0].Pitch = gs_apCtrlOption[1].Pitch = gs_apCtrlOption[2].Pitch = gs_apCtrlOption[3].Pitch = 45;
	gs_apCtrlOption[0].Roll = 0;
	gs_apCtrlOption[1].Roll = 90;
	gs_apCtrlOption[2].Roll = 180;
	gs_apCtrlOption[3].Roll = 270;
#endif  // _MULTI_VIEW

	// Set the input/output picture information (YUV420P)
	{
		DWORD dwSrcBufferSize, dwDstBufferSize;
		BYTE *pbySrcBuffer = NULL, *pbyDstBuffer = NULL;		

		option.InVPicture.Width = option.InVPicture.Height = option.InVPicture.Stride = dwFovSize;
		option.InVPicture.Format = FE_PIXELFORMAT_YUV420P;

		option.OutVPicture.Width = option.OutVPicture.Stride = pOverlay->w;
		option.OutVPicture.Height = pOverlay->h;
		option.OutVPicture.Format = FE_PIXELFORMAT_YUV420P;		
		
		dwSrcBufferSize = option.InVPicture.Stride * option.InVPicture.Height * 3 / 2;
		
		if (NULL == (pbySrcBuffer = (BYTE *)malloc(dwSrcBufferSize)))
		{
			return -1;
		}

		memset(pbySrcBuffer, 0, dwSrcBufferSize);
		option.InVPicture.Buffer = pbySrcBuffer;		

		fread(option.InVPicture.Buffer, dwSrcBufferSize, 1, pfSample);
		fclose(pfSample);

		dwDstBufferSize = option.OutVPicture.Stride * option.OutVPicture.Height * 3 / 2;  

		if (NULL == (pbyDstBuffer = (BYTE *)malloc(dwDstBufferSize)))
		{
			return -1;
		}

		option.OutVPicture.Buffer = pbyDstBuffer;
	}

	option.FOVCenter.X = dwFovSize >> 1;
	option.FOVCenter.Y = dwFovSize >> 1;
	option.FOVRadius = dwFovSize >> 1; 

	switch (mountType)
	{
	case 0:
		option.MountType = FE_MOUNT_WALL;
		break;
	case 1:
		option.MountType = FE_MOUNT_CEILING;
		break;
	case 2:
		option.MountType = FE_MOUNT_FLOOR;
		break;
	default:
		break;
	}

	option.DewarpType = FE_DEWARP_RECTILINEAR;


	option.OutRoi.Left = 0;
	option.OutRoi.Top = 0;
	option.OutRoi.Right = pOverlay->w;
	option.OutRoi.Bottom = pOverlay->h;

#ifdef _MULTI_VIEW
	gs_apCtrlOption[0].OutRoi.Left = option.OutRoi.Right >> 1;
	gs_apCtrlOption[0].OutRoi.Right = option.OutRoi.Right;
	gs_apCtrlOption[0].OutRoi.Bottom = option.OutRoi.Bottom >> 1;
	gs_apCtrlOption[1].OutRoi.Right = option.OutRoi.Right >> 1;
	gs_apCtrlOption[1].OutRoi.Bottom = option.OutRoi.Bottom >> 1;
	gs_apCtrlOption[2].OutRoi.Top = option.OutRoi.Bottom >> 1;
	gs_apCtrlOption[2].OutRoi.Right = option.OutRoi.Right >> 1;
	gs_apCtrlOption[2].OutRoi.Bottom = option.OutRoi.Bottom;
	gs_apCtrlOption[3].OutRoi.Left = option.OutRoi.Right >> 1;
	gs_apCtrlOption[3].OutRoi.Top = option.OutRoi.Bottom >> 1;
	gs_apCtrlOption[3].OutRoi.Right = option.OutRoi.Right;
	gs_apCtrlOption[3].OutRoi.Bottom = option.OutRoi.Bottom;
#endif  // _MULTI_VIEW

	option.Flags = (FE_OPTION_INIMAGEHEADER | FE_OPTION_INIMAGEBUFFER | FE_OPTION_OUTIMAGEHEADER | FE_OPTION_OUTIMAGEBUFFER | 
		            FE_OPTION_FOVCENTER | FE_OPTION_FOVRADIUS | FE_OPTION_MOUNTTYPE | FE_OPTION_DEWARPTYPE | FE_OPTION_OUTROI);

	UpdateOutVPicture(hFECtx, &option, pOverlay);

	EventLoop(hFECtx, &option, pOverlay);

	if (option.OutVPicture.Buffer)
	{
		free(option.OutVPicture.Buffer);
	}

	if (option.InVPicture.Buffer)
	{
		free(option.InVPicture.Buffer);
	}

	// Release Fisheye library
	Fisheye_Release(&hFECtx);

	SDL_FreeYUVOverlay(pOverlay);
	
	SDL_Quit();	

	return 0;
}

void EventLoop(HANDLE hObject, FEOPTION *pOption, SDL_Overlay *pOverlay)
{
	int bExit = 0;
	SDL_Event evt;

	pOption->Flags = 0;
	
	while (!bExit)
	{	
		SDL_WaitEvent(&evt);
		
		switch (evt.type)
		{
		case SDL_KEYDOWN:
			switch (evt.key.keysym.sym)
			{
			case SDLK_UP:
				pOption->Pitch += 1;
				pOption->Flags |= FE_OPTION_PITCH;
#ifdef _MULTI_VIEW
				{
					int k;
					for (k = 0; k < MULTI_VIEW_COUNT; ++k)
					{
						gs_apCtrlOption[k].Pitch += 1;
					}
				}
#endif
				break;
			case SDLK_DOWN:
				pOption->Pitch -= 1;
				pOption->Flags |= FE_OPTION_PITCH;
#ifdef _MULTI_VIEW
				{
					int k;
					for (k = 0; k < MULTI_VIEW_COUNT; ++k)
					{
						gs_apCtrlOption[k].Pitch -= 1;
					}
				}
#endif
				break;
			case SDLK_LEFT:
				if (FE_MOUNT_WALL == pOption->MountType)
				{
					pOption->Yaw -= 1;
					pOption->Flags |= FE_OPTION_YAW;
				}
				else
				{
					if ((pOption->DewarpType == FE_DEWARP_FULLVIEWPANORAMA) || (pOption->DewarpType == FE_DEWARP_DUALVIEWPANORAMA))
					{
						pOption->Scroll -= 1;
						pOption->Flags |= FE_OPTION_SCROLL;
					}
					else
					{					
						pOption->Roll -= 1;
						pOption->Flags |= FE_OPTION_ROLL;

#ifdef _MULTI_VIEW
						{
							int k;
							for (k = 0; k < MULTI_VIEW_COUNT; ++k)
							{
								gs_apCtrlOption[k].Roll -= 1;
							}
						}
#endif
					}
				}
				break;
			case SDLK_RIGHT:
				if (FE_MOUNT_WALL == pOption->MountType)
				{
					pOption->Yaw += 1;
					pOption->Flags |= FE_OPTION_YAW;
				}
				else
				{
					if ((pOption->DewarpType == FE_DEWARP_FULLVIEWPANORAMA) || (pOption->DewarpType == FE_DEWARP_DUALVIEWPANORAMA))
					{
						pOption->Scroll += 1;
						pOption->Flags |= FE_OPTION_SCROLL;
					}
					else
					{
						pOption->Roll += 1;
						pOption->Flags |= FE_OPTION_ROLL;
#ifdef _MULTI_VIEW
						{
							int k;
							for (k = 0; k < MULTI_VIEW_COUNT; ++k)
							{
								gs_apCtrlOption[k].Roll += 1;
							}
						}
#endif
					}
				}
				break;
			case SDLK_KP_PLUS:
				pOption->Scale += 1;
				pOption->Flags |= FE_OPTION_SCALE;
				break;
			case SDLK_KP_MINUS:
				pOption->Scale -= 1;
				pOption->Flags |= FE_OPTION_SCALE;
			default:
				break;
			}
			UpdateOutVPicture(hObject, pOption, pOverlay);
			break;
		case SDL_QUIT:
			bExit = 1;
			break;
		default:
			break;
		}	
	}
}

SCODE UpdateOutVPicture(HANDLE hObject, FEOPTION *pOption, SDL_Overlay *pOverlay)
{
	SCODE scRet;
	int sdlRet;
	SDL_Rect roi;

	scRet = Fisheye_SetOption(hObject, pOption);
	if (S_OK != scRet)
	{
		return scRet;
	}

	scRet = Fisheye_OneFrame(hObject);
	if (S_OK != scRet)
	{
		return scRet;
	}

#ifdef _MULTI_VIEW
	{
		FEOPTION tOption;
		int k;
		
		memset(&tOption, 0, sizeof(FEOPTION));
		
		for (k = 0; k < MULTI_VIEW_COUNT; ++k)
		{
			tOption.Pitch = gs_apCtrlOption[k].Pitch;
			tOption.Roll = gs_apCtrlOption[k].Roll;
			memcpy(&tOption.OutRoi, &gs_apCtrlOption[k].OutRoi, sizeof(FERECT));

			tOption.Flags = (FE_OPTION_PITCH | FE_OPTION_ROLL | FE_OPTION_OUTROI);
			Fisheye_SetOption(hObject, &tOption);
			Fisheye_OneFrame(hObject);
		}
	}
#endif  // _MULTI_VIEW

	// Write the output raw data to file
	{
		FILE *pStream = fopen("sample_out.yuv", "wb");
		fwrite(pOption->OutVPicture.Buffer, (pOption->OutVPicture.Stride * pOption->OutVPicture.Height * 3) >> 1, 1, pStream);
		fclose(pStream);
	}

	if (0 == (sdlRet = SDL_LockYUVOverlay(pOverlay)))
	{
		int offset1, offset2, width, height, i;
		BYTE *buffer[3], *src, *dst;
		int stride[3];

		offset1 = pOption->OutVPicture.Stride * pOption->OutVPicture.Height;
		offset2 = offset1 * 5 >> 2;
		width = pOption->OutVPicture.Width;
		height = pOption->OutVPicture.Height;

		buffer[0] = pOption->OutVPicture.Buffer;
		buffer[1] = pOption->OutVPicture.Buffer + offset1;
		buffer[2] = pOption->OutVPicture.Buffer + offset2;

		stride[0] = pOption->OutVPicture.Stride;
		stride[2] = stride[1] = pOption->OutVPicture.Stride >> 1;

		// Copy luminance component (Y)
		src = buffer[0];
		dst = (BYTE*)pOverlay->pixels[0];
		for (i = 0; i < height; ++i, src += stride[0], dst += pOverlay->pitches[0])
		{
			memcpy(dst, src, width);
		}

		// Copy chroma component (U, V)
		width >>= 1;
		height >>= 1;

		src = buffer[1];
		dst = (BYTE*)pOverlay->pixels[2];
		for (i = 0; i < height; ++i, src += stride[1], dst += pOverlay->pitches[1])
		{
			memcpy(dst, src, width);
		}

		src = buffer[2];
		dst = (BYTE*)pOverlay->pixels[1];
		for (i = 0; i < height; ++i, src += stride[2], dst += pOverlay->pitches[2])
		{
			memcpy(dst, src, width);
		}

		SDL_UnlockYUVOverlay(pOverlay);
	}

	roi.x = 0;
	roi.y = 0;
	roi.w = pOverlay->w;
	roi.h = pOverlay->h;
	
	SDL_DisplayYUVOverlay(pOverlay, &roi);

	return S_OK;
}
